/*
 * DbApplication.java
 *
 * Rewritten for Hibernate 2007-08-24 Christoph Knabe
 * Created on 28. Mai 2001, 14:51 by Christoph Knabe
 */

package bank3tier.db;

import javax.persistence.EntityManager;
import javax.persistence.EntityTransaction;

import org.hibernate.Session;
import org.hibernate.Transaction;


/**Here the database access layer is composed.
 * The persistence manager is elected, as well as the necessary database tables.
 * This DbApplication uses _HibernatePool as persistence layer and the tables
 * Kunde, Konto, Verwaltet. 
 * For efficiency this class should be used as a singleton.
* @author  2005-03-29  Christoph Knabe
*/
public class DbApplication {

    /**The object of the persistence manager*/
    /*package*/ _HibernatePool _pool;
        
    /** Creates new DbBankApplication */
    public DbApplication(final Class[] i_classes) {
        try{
            this._pool = new _HibernatePool(i_classes);
            //bank3tier.db.TestAll.setUpEmptyDB();
            //DbObject.dbOpenDB(this);
            //_openDB();
		    //_createTablesBySql(DbObject.oidSuffix); dbEndTransaction(true);  //done by hbm2ddl.auto in hibernate.cfg.xml
		}catch(final Exception ex){
		    throw new multex.Failure(
                "Failure creating the application specific database", ex
		    );
		}
    }

    /**Creates the application specific database tables
     //Not necessary for Hibernate, as this has schema export to database.
    private void _createTablesBySql(final String i_oidSuffix) {
        _pool.executeSql("create table Kunde ("
            +" oid INTEGER NOT NULL PRIMARY KEY"
            +",db__updateVersion INTEGER NOT NULL"
            +",name VARCHAR(50) NOT NULL"
            +",geburt DATE NOT NULL"
            +" )"
        );
        _pool.executeSql("create table Konto ("
            +" oid INTEGER NOT NULL PRIMARY KEY"
            +",db__updateVersion INTEGER NOT NULL"
            +",saldo INTEGER NOT NULL"
            +",eroeffnet DATE NOT NULL"
            +",geschlossen DATE"
            +" )"
        );
        _pool.executeSql("create table Verwaltet ("
            +" oid INTEGER NOT NULL PRIMARY KEY"
            +",db__updateVersion INTEGER NOT NULL"
            +",kunde" + i_oidSuffix + " INTEGER NOT NULL"
            +",konto" + i_oidSuffix + " INTEGER NOT NULL"
            +",inhaber BOOLEAN NOT NULL"
            +" )"
        );
    }*/
    
    /**Deletes the application specific database tables, if they exist.
     //Not necessary with Hibernate, as this uses schema export to database.
    private void _deleteTables() {
        _pool.executeSql("drop table Kunde if exists");
        _pool.executeSql("drop table Konto if exists");
        _pool.executeSql("drop table Verwaltet if exists");
    }*/
    
    /**Gets the Pool interface for this persistence layer*/
    public Pool getPool() {
        return this._pool;
    }

    /** Mark the beginning of an transaction.
      Unconfirms all modifications made to the database after the last
      call to {@link #endTransaction(boolean)}.
     */
    public void dbBeginTransaction() { 
        this._pool.getEntityManager().getTransaction().begin();
    }

    /** Confirm or unconfirm the modifications made to the database in the
      actual transaction (since the last commit) and close this transaction.
      @param i_ok true if the transaction succeeded, then the modifications
        will be confirmed, otherwise unconfirmed.
     */
    public void dbEndTransaction(final boolean i_ok) {
        final EntityManager session = this._pool.getEntityManager();
        final EntityTransaction transaction = session.getTransaction();
        if(i_ok){
            transaction.commit();
        }else{
            session.flush();
            transaction.rollback();
        }
        //session.close();  //Bei Hibernate-ThreadLocal-Session anscheinend schon bei TA-Ende geschehen. 07-08-24
        //Ohne session.close(), um denselben EntityManager mehrmals zu verwenden
    }
    

}