package bank3tier.db; //_HibernatePool.java

//2005-04-07  Knabe  Erstellt


import bank3tier.db.DbObject.OwnerPoolChangedExc;
import bank3tier.db.Pool.LinkFailure;
import bank3tier.db.Pool.NoPersistentClassExc;
import bank3tier.db.Pool.UnexpectedExampleClassExc;
import multex.Exc;
import multex.Failure;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Modifier;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.persistence.EntityManager;
import javax.persistence.EntityManagerFactory;
import javax.persistence.Persistence;
import javax.persistence.Query;

import org.hibernate.HibernateException;
import org.hibernate.event.PostLoadEvent;
import org.hibernate.event.def.DefaultPostLoadEventListener;



/**Implementation of the persistence layer by Hibernate-JPA. Holds one database session.
 * @author knabe 2007-08-13
 */
/*package*/ class _HibernatePool implements Pool, PoolAdmin {


    private final Class[] persistentClasses;
    private final /*SessionFactory*/ EntityManagerFactory sessionFactory;
    private final EntityManager entityManager; //Should be ThreadLocal!

    /*static {
        try {
            // Create the SessionFactory 
            sessionFactory = new Configuration().configure().buildSessionFactory();
        } catch (final Throwable ex) {
            throw new RuntimeException("Initial SessionFactory creation failed.", ex);
        }
    }*/

    public /*SessionFactory*/ EntityManagerFactory getSessionFactory() {
        return this.sessionFactory;
    }
    
    /**Opens an application specific Hibernate database.
     * */
    public _HibernatePool(final Class[] persistentClasses){
        this.persistentClasses = persistentClasses;
        try {
            //Hibernate:
            //final Configuration configuration = new Configuration(); //from hibernate.cfg.xml
            //final AnnotationConfiguration configuration = new AnnotationConfiguration(); //from annotations
            
            /*final Ejb3Configuration configuration = new Ejb3Configuration(); //from Hibernate Entity Manager
            for (final Class aClass: persistentClasses) {
                _checkPoolConstructorPrerequisites(aClass);
                _checkHibernateConstructorPrerequisites(aClass);
                configuration//.addClass(aClass);
                .addAnnotatedClass(aClass);
            }*/
            
            //final Configuration config = configuration;

            // Apply this event listener (programmatically)
            //TODO Umstellen auf lifecycle callback method gemäß EJB3:
            //configuration.setListener("post-load",new NoisyPostLoadEventListener());
            //Wahrscheinlich unnötig:
            //configuration.configure();
            
            //this.sessionFactory = configuration.buildSessionFactory();
            //this.sessionFactory = configuration.buildEntityManagerFactory();
            this.sessionFactory = Persistence.createEntityManagerFactory("bank3tier");
            this.entityManager = this.sessionFactory.createEntityManager();
		} catch (final Exception ex) {
            throw new Failure("Could not create persistence", ex);
		}
    }

    private void _checkHibernateConstructorPrerequisites(final Class aClass) throws InstantiationException, IllegalAccessException {
        final Constructor parameterlessConstructor;
        try{ parameterlessConstructor = aClass.getDeclaredConstructor();        
        }catch(Exception ex){
            throw new Failure("Persistent class {0} must have parameterless (non-private) constructor", ex, aClass.getName());
        }
        final int modifiers = parameterlessConstructor.getModifiers();
        if(Modifier.isPrivate(modifiers)){
            throw new Failure("The parameterless constructor of persistent class {0} must be non-private", null, aClass.getName());
        }
    }
    
    private void _checkPoolConstructorPrerequisites(final Class aClass){
        if(!DbObject.class.isAssignableFrom(aClass)){
            throw new Failure("Persistent class {0} must be a subclass of {1}", null, aClass.getName(), DbObject.class.getName());
        }
    }
    
    @Override
    public void finalize(){
        this.entityManager.close();
        this.sessionFactory.close();
    }
    public void close(){
    }

    //implements from Pool
    public void insert(final DbObject io_object) throws OwnerPoolChangedExc{
        //final  EntityManager entityManager = getEntityManager();
        io_object.setPool(this);
        this.entityManager.persist(io_object);
    }
    
//  implements from Pool
    public void reattach(final DbObject io_object){
        final  EntityManager session = getEntityManager();
        session.merge(io_object);
    }
    
    //implements from Pool
    public boolean save(final DbObject io_object) throws ConcurrentUpdateExc, DbObject.OwnerPoolChangedExc {
        final  EntityManager session = getEntityManager();
        final boolean isPersistent = session.contains(io_object);
        if (!isPersistent) {
            io_object.setPool(this);
        }        
        session.merge(io_object);
		return !isPersistent;
    }

    /**Liefert den aktuellen EntityManager.
     */
    /*package*/ EntityManager getEntityManager(){
        return this.entityManager;
    }

    //implements from Pool
    public final boolean delete(final DbObject io_object)
    { 
        final  EntityManager session = getEntityManager();
        final boolean isPersistent = session.contains(io_object);
        session.remove(io_object);
        io_object.oid = Pool.deletedOid;
        return isPersistent;
    }

    public <T extends DbObject> T find(final Class<T> i_persistentClass, final Integer i_oid)
    throws NoPersistentClassExc, Pool.OidNotFoundExc {
        _checkPersistentClass(i_persistentClass);

        final  EntityManager session = getEntityManager();
        final Object object = session.find(i_persistentClass, i_oid);
        if(null==object){
            throw new Pool.OidNotFoundExc(i_oid, i_persistentClass.getSimpleName());
        }
        return (T)object;
    }
    
    public <T extends DbObject> T findOneBykey(final Class<T> i_persistentClass, final String i_keyFieldName, final Object i_keyFieldValue) 
    throws NoPersistentClassExc, KeyNotFoundExc, KeyNotUniqueExc {
        final List<T> results = findManyByKey(i_persistentClass, i_keyFieldName, i_keyFieldValue);
        final int size = results.size();
        if(size<1){
            throw new KeyNotFoundExc(i_persistentClass, i_keyFieldName, i_keyFieldValue);
        }
        if(size>1){
            throw new KeyNotUniqueExc(i_persistentClass, i_keyFieldName, i_keyFieldValue, size);
        }
        return results.get(0);
    }

    //implements from Pool
    public <T extends DbObject> List<T> findAll(final Class<T> i_persistentClass) throws NoPersistentClassExc {
        _checkPersistentClass(i_persistentClass);
        try {
            final  EntityManager session = getEntityManager();
            final Query q = session.createQuery("select o from " + i_persistentClass.getName() + " o ORDER BY oid DESC");
            final List<T> results = q.getResultList();
            return results;
        }catch(Exception ex){throw new multex.Failure(
            "Failure searching all objects of class {0}", ex, i_persistentClass.getName());
        }
    }
    
    public <T extends DbObject> List<T> findManyByKey(final Class<T> i_persistentClass, final String i_keyFieldName, final Object i_keyFieldValue) throws NoPersistentClassExc {
        _checkPersistentClass(i_persistentClass);
        final String className = i_persistentClass.getName();
        final  EntityManager entityManager = getEntityManager();
        final Query q = entityManager.createQuery("select o from " + className + " o where " + i_keyFieldName + " = :keyValue ORDER BY o.oid DESC");
        q.setParameter("keyValue", i_keyFieldValue);
        final List<T> results = q.getResultList();
        return results;
    }
    

	//implements from Pool
    public boolean areLinked(final DbObject i_left, final Class<? extends DbObject> i_linkClass, final DbObject i_right) throws NoPersistentClassExc {
        _checkPersistentClass(i_linkClass);
		
        final EntityManager session = getEntityManager();
        final Query jpaQuery = session.createQuery("select lo from " + i_linkClass.getName() + " lo "
                +"WHERE " + _linkCondition(i_left, "left")
                + " AND " + _linkCondition(i_right, "right"));
        jpaQuery.setParameter("left", i_left.getOid());
        jpaQuery.setParameter("right", i_right.getOid());
        final List linkObjects = jpaQuery.getResultList();
        return linkObjects.size() > 0;
    }

    //implements for Pool
    public <T extends DbObject> boolean hasLinks(final Class<T> i_resultClass, final DbObject i_object) throws LinkFailure, NoPersistentClassExc {
        //TODO K�nnte optimiert werden mit einer HQL-Aggregatfunktion wie COUNT(...), damit die Verkn�pfungspartner nicht wirklich gelesen werden m�ssen. 07-08-29
        return getLinks(i_resultClass, i_object).size() > 0;
    }
    
    //implements for Pool
    public <T extends DbObject> List<T> getLinks(final Class<T> i_resultClass, final DbObject i_object) throws LinkFailure, NoPersistentClassExc{
        _checkPersistentClass(i_resultClass);
        final EntityManager session = getEntityManager();
        final Query jpaQuery = session.createQuery("select lo from " + i_resultClass.getName() + " lo "
                +"WHERE " + _linkCondition(i_object, "left"));
        jpaQuery.setParameter("left", i_object.getOid());
        final List<T> result = jpaQuery.getResultList();
        return result;
    }
  
    //implements for Pool
    public <T extends DbObject> List<T> findManyByQuery(
        final Class<T> i_resultClass, 
        final Class i_queryClass, 
        final Object[] i_args
    )
    throws NoPersistentClassExc, NoQueryClassExc, ArgumentCountExc, ArgumentTypeExc
    {
        _checkPersistentClass(i_resultClass); //throws NoPersistentClassExc
        if(!HibernateQuery.class.isAssignableFrom(i_queryClass)){
            throw new NoQueryClassExc(i_queryClass, HibernateQuery.class);
        }
        try {
            final HibernateQuery hibernateQuery = (HibernateQuery)i_queryClass.newInstance();
            final List<T> results = hibernateQuery.execute(this, i_resultClass, i_args);
            return results;
        }catch(ArgumentCountExc ex){throw ex;
        }catch(ArgumentTypeExc ex){throw ex;
        }catch(Exception ex){
            final List<Object> argsAsList = i_args!=null && i_args.length>0 ? Arrays.asList(i_args) : null;
            throw new multex.Failure(
                    "Failure searching objects of class {0} by query {1} with arguments {3}"
                    , ex, i_resultClass.getName(), i_queryClass.getName(), argsAsList
            );
        }
    }
    
	/**Returns the SQL condition, that i_linkPartner's OID is the same as its table name followed by the OID suffix.
	 * @param side The side ("left" or "right")
     * @throws NoPersistentClassExc*/
	private String _linkCondition(final DbObject i_linkPartner, final String side){
        final Class<? extends DbObject> partnerClass = i_linkPartner.getClass(); //e.g.  LgKunde
        final String linkName = _computeLinkName(partnerClass); //e.g. kunde
		return linkName + DbObject.oidSuffix + " = :" + side;
	}

    private <T extends DbObject> void _checkPersistentClass(final Class<T> i_persistentClass) throws NoPersistentClassExc {
        if(!DbObject.class.isAssignableFrom(i_persistentClass)){
            throw new NoPersistentClassExc(i_persistentClass);
        }
        final List<Class> persistentClassesList = Arrays.asList(this.persistentClasses);
        if(!persistentClassesList.contains(i_persistentClass)){
            throw new NoPersistentClassExc(i_persistentClass);
        }
	}
    
    /**Strips leading "Lg" from the simple class name and uncapitalizes the first letter.
     * @throws Exc Partner class name does not start with prefix {0}
     */
    private String _computeLinkName(final Class partnerClass) throws Exc {
        final String partnerClassSimpleName = partnerClass.getSimpleName();
        final String lgPrefix = "Lg";
        if(!partnerClassSimpleName.startsWith(lgPrefix)){
            throw new Exc("Partner class name does not start with prefix {0}", lgPrefix);
        }
        final StringBuffer buf = new StringBuffer(partnerClassSimpleName);
        buf.delete(0, 2);
        buf.setCharAt(0, Character.toLowerCase(buf.charAt(0)));
        return buf.toString();
    }

    
}
